
/*
 *  source - $Id: internal.c,v 1.4 1997/06/11 21:39:52 qian Exp $
 *
 *  internel API of the memory-based guard tool
 */

#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/swap.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/pgtable.h>
#include <asm/string.h>

#include <asm/memguard_machdep.h>
#include <memguard/macros.h>
#include <memguard/types.h>

#include <memguard/internal.h>
#include <memguard/memguard_internal.h>

#include <asm/memguard.h>
/*
#include <memguard/main.h>
*/


/*********************************************************************
variables (public)
*********************************************************************/

/*
 * Variables  : int mgEnabled
 * Description: if the memory-based guard tool is enabled
 */
/* if the memory-based guard tool is enabled */
int mgEnabled;

/*
 * Variables  : mgNinvariant
 *              mgInvariants
 *              mgInvariantsFree
 * Description: information of invariants
 */
/* the number of control blocks of invariants */
u_int mgNinvariant;

/* an allocated array of control blocks of invariants */
mginvariant_t *mgInvariants;

/* a free list of control blocks of invariants */
mginvariant_t *mgInvariantsFree;

/*
 * Variables  : mgNregion
 *              mgRegions
 *              mgIndexRegion
 * Description: information of guarded regions
 */
/* the number of control blocks of regions */
u_int mgNregion;

/* an allocated array of control blocks of regions */
mgregion_t *mgRegions;

/* the current free index of the control blocks of guarded regions */
u_int mgIndexRegion;

/*
 * Variables  : mgErrno
 * Description: the error number for the internal use
 */
/* the error number for the internal use */
int mgErrno;


/*********************************************************************
variables (private)
*********************************************************************/


/*********************************************************************
function prototypes (private)
*********************************************************************/

static int   mg_addregion(caddr_t svaddr, caddr_t evaddr);
#    define ADDREG_OK		0	/* success */
#    define ADDREG_FULL		-1	/* too many regions */
#    define ADDREG_NOSPACE	-2	/* too many pages */


/*********************************************************************
functions (public)
*********************************************************************/

/*
 * Function   : mgInit
 * Description: initialize data structures
 * Parameters : benabled   - if the tool is enabled initially
 *              nregion    - number of guarded regions
 *              ninvariant - number of invariants
 * Returns    : return flag (int)
 *              MG_INIT_OK		success
 *              MG_INIT_NREGION		invalid number of regions
 *              MG_INIT_NINVARIANT	invalid number of invariants
 * Plan       : 1. initialize information of regions and invariants
 *              2. set initial tool state
 *              3. allocate control blocks of regions
 *              4. allocate control blocks of invariants
 *              5. reset these data structures
 *              6. return ok
 * Note       : must called before anything else
 */
int
mgInit(int benabled, u_int nregion, u_int ninvariant)
{
    /*
     * 1. initialize information of regions and invariants
     */
    mgNregion = 0;
    mgRegions = NULL;
    mgIndexRegion = 0;

    mgNinvariant = 0;
    mgInvariants = NULL;
    mgInvariantsFree = NULL;

    /*
     * 2. set initial tool state
     */
    mgEnabled = benabled;

    /*
     * 3. allocate control blocks of regions
     */
    if ((mgRegions = (mgregion_t *)
	    mgAllocate(nregion, sizeof(mgregion_t))) == NULL) {
        /* allocate failed, return error */
        return(MG_INIT_NREGION);
    }
    /* success */
    mgNregion = nregion;

    /*
     * 4. allocate control blocks of invariants
     */
    if ((mgInvariants = (mginvariant_t *)
	    mgAllocate(ninvariant, sizeof(mginvariant_t))) == NULL) {
        /* allocate failed, return error */
        return(MG_INIT_NINVARIANT);
    }
    /* success */
    mgNinvariant = ninvariant;

    /*
     * 5. reset these data structures
     */
    mgResetDS();

    /*
     * 6. return ok
     */
    return(MG_INIT_OK);
}

/*
 * Function   : mgAddRegion
 * Description: add a guarded region
 * Parameters : vaddr - the start virtual address of the region
 *              size  - the size of the region
 * Returns    : return flag (int)
 *              MG_ADDREG_OK		success
 *              MG_ADDREG_VADDR		invalid virtual address
 *              MG_ADDREG_SIZE		invalid region size
 *              MG_ADDREG_FULL		too many regions
 *              MG_ADDREG_NOSPACE	too many pages
 * Plan       : 1. check virtual address
 *              2. check region size
 *              3. add this region, return
 * Note       : none
 */
int
mgAddRegion(caddr_t vaddr, u_int size)
{
    register caddr_t svaddr_aligned, evaddr_aligned;

    /*
     * 1. check virtual address
     */
    if (vaddr <= (caddr_t)0) {
	/* invalid virtual address, return error */
	return(MG_ADDREG_VADDR);
    }
    svaddr_aligned = mgAlignStartAddr(vaddr);

    /*
     * 2. check region size
     */
    if (size <= 0) {
	/* invalid region size, return error */
	return(MG_ADDREG_SIZE);
    }
    evaddr_aligned = mgAlignEndAddr(vaddr + size);

    /*
     * 3. add this region, return
     */
    switch (mg_addregion(svaddr_aligned, evaddr_aligned)) {
    case ADDREG_OK :		/* success */
	/* return success */
	return(MG_ADDREG_OK);
	/* return, no break */
    case ADDREG_FULL :		/* too many regions */
	/* return error */
	return(MG_ADDREG_FULL);
	/* return, no break */
    case ADDREG_NOSPACE :	/* too many pages */
	/* return error */
	return(MG_ADDREG_NOSPACE);
	/* return, no break */
    default :			/* impossible */
        /* return error */
        mgReportImpossible("mgAddRegion()");
        return(MG_ADDREG_FULL);
	/* return, no break */
    }
}

/*
 * Function   : mgDeleteRegion
 * Description: delete a guarded region
 * Parameters : vaddr - the start virtual address of the region
 * Returns    : return flag (int)
 *              MG_DELREG_OK		success
 *              MG_DELREG_VADDR		invalid virtual address
 * Plan       : 1. return error, do NOT support
 * Note       : none
 */
int
mgDeleteRegion(caddr_t vaddr)
{
    /*
     * 1. return error, do NOT support
     */
    return(MG_DELREG_VADDR);
}

/*
 * Function   : mgFault
 * Description: memguard page protection fault function
 * Parameters : vaddr   - virtual address
 *              length  - writing length in the page
 * Returns    : return flag (int)
 *              MG_FAULT_YES	this is a memguard fault
 *              MG_FAULT_NO	this is not a memguard fault
 * Plan       : 1. get the corresponding page
 *              2. check if this page is invalid
 *              3. report memguard fault
 *              4. check original protection
 *              5. search invariants of the page
 *              6. return yes
 * Note       : none
 */
int
mgFault(caddr_t vaddr, u_short length)
{
    register mgpage_t      *pgp;
    register mginvariant_t *pinvariant;
    register u_short        offset;
    register u_short        offset_invariant;

    /*
     * 1. get the corresponding page
     */
    if ((pgp = mgGetPage(vaddr, length)) == NULL) {
	/* not found, return no */
        return(MG_FAULT_NO);
    }
    /* found */

    /*
     * 2. check if this page is invalid
     */
    if ((pinvariant = pgp->gp_invariants) == NULL) {
	/* invalid page, return no */
        return(MG_FAULT_NO);
    }

    /*
     * 3. report memguard fault
     */
#ifdef MG_DEBUG
    mgReportFault(vaddr, length);
#endif

    /*
     * 4. check original protection
     */
    if (! pgp->gp_prot) {
	/* original protection fail, return no */
	return(MG_FAULT_NO);
    }
    /* original protection ok */

    /*
     * 5. search invariants of the page
     */
    offset = mgAddr2Offset(vaddr);
    for (/* pinvariant = pgp->gp_invariants */; pinvariant != NULL;
	    pinvariant = pinvariant->gi_next) {
	/* check no conflict */
	if (offset + length <= (offset_invariant = pinvariant->gi_offset)
	        || offset >= offset_invariant + pinvariant->gi_length) {
            /* no conflict, continue */
	    continue;
        }
	/* conflict */

        /* report error, break (do not return no) */
	mgReportViolate(pinvariant, vaddr, offset, length);
	break;
	/* do not return no
	return(MG_FAULT_NO);
	*/
    }

    /*
     * 6. return success
     */
    return(MG_FAULT_YES);
}

/*
 * Function   : mgResetDS
 * Description: reset data structures
 * Parameters : none
 * Returns    : none
 * Plan       : 1. reset control blocks of region and corresponding
 *                 control blocks of page
 *              2. reset control blocks of invariant
 * Note       : the tool must be initialized and disabled
 *              the process cannot be interrupted because the data
 *                structures are inconsistent during the process
 */
void
mgResetDS()
{
    register int ix;

    /*
     * 1. reset control blocks of region and corresponding
     *    control blocks of page
     */
    for (ix = 0; ix < mgIndexRegion; ix ++) {
	register mgpage_t *pgp, *pgp_tail;

        /* NOTE: do not reset protection field */
        for (pgp_tail = (pgp = mgRegions[ix].gr_pages)
		+ mgRegions[ix].gr_npage; pgp < pgp_tail; pgp ++) {
            pgp->gp_invariants = NULL;
	}
    }

    /*
     * 2. reset control blocks of invariant
     */
    mgInvariantsFree = NULL;
    for (ix = 0; ix < mgNinvariant; ix ++) {
	mgInvariants[ix].gi_next = mgInvariantsFree;
	mgInvariantsFree = &mgInvariants[ix];
    }
}

/*
 * Function   : mgGetPage
 * Description: get page with virtual address in
 * Parameters : vaddr - invariant virtual address
 *              size  - invariant size
 * Returns    : pointer to the found page (mgpage_t *)
 *              != NULL	pointer to the found page
 *              == NULL	not found
 * Plan       : 1. search regions
 *              2. not found, set error, return NULL
 * Note       : none
 */
mgpage_t *
mgGetPage(caddr_t vaddr, u_int size)
{
    register mgregion_t *pregion;

    /*
     * 1. search regions
     */
    for (pregion = &mgRegions[mgIndexRegion - 1];
	    pregion >= &mgRegions[0]; pregion --) {
        /* check region */
        if (vaddr >= pregion->gr_svaddr && vaddr < pregion->gr_evaddr) {
	    /* virtual address in this region, check size */
	    if (vaddr + size <= pregion->gr_evaddr) {
		/* ok, return the corresponding page */
	        return(&pregion->gr_pages[
		        mgAddr2Pageno(vaddr - pregion->gr_svaddr)]);
	    } else {
		/* invalid size, set error, return NULL */
		mgErrno = MG_GETPAGE_SIZE;
		return(NULL);
	    }
	}
    }
    /* not found */

    /*
     * 2. not found, set error, return NULL
     */
    mgErrno = MG_GETPAGE_VADDR;
    return(NULL);
}


/*********************************************************************
functions (private)
*********************************************************************/

/*
 * Function   : mg_addregion
 * Description: add a guarded region (recursive version)
 * Parameters : svaddr - start virtual address (aligned)
 *              evaddr - (EXCLUSIVE) end virtual address (aligned)
 * Returns    : return flag (int)
 *              ADDREG_OK	success
 *              ADDREG_FULL	too many regions
 *              ADDREG_NOSPACE	too many pages
 * Plan       : 1. search regions to adjust this region
 *              2. check free regions
 *              3. allocate free pages
 *              4. allocate the region
 *              5. initialize the region
 *              6. initialize pages
 *              7. return ok
 * Note       : none
 */
static int
mg_addregion(caddr_t svaddr, caddr_t evaddr)
{
    u_int npage;
    register mgpage_t   *pgp;
    register mgregion_t *pregion;
    register int ix;

    /*
     * 1. search regions to adjust this region
     */
    for (pregion = &mgRegions[mgIndexRegion - 1];
	    pregion >= &mgRegions[0]; pregion --) {
        caddr_t svaddr_old, evaddr_old;
        int return_val;

	svaddr_old = pregion->gr_svaddr;
	evaddr_old = pregion->gr_evaddr;

        /* check region */
	if (evaddr <= svaddr_old) {
	    /* entirely before the original region, do nothing */
	    continue;
	}
	/* evaddr > svaddr_old */
	if (svaddr >= evaddr_old) {
	    /* entirely after the original region, do nothing */
	    continue;
	}
	/* evaddr > svaddr_old && svaddr < evaddr_old */
	if (svaddr < svaddr_old) {
	    /* svaddr < svaddr_old < evaddr */
	    if (evaddr <= evaddr_old) {
		/* overlay from left, truncate right portion */
		evaddr = svaddr_old;
		continue;
	    } else {
		/* encircling the original region, add two regions */
		if ((return_val = mg_addregion(svaddr, svaddr_old))
			!= ADDREG_OK) {
                    /* failed, return it */
		    return(return_val);
		}
                /* success, add next region */
		if ((return_val = mg_addregion(evaddr_old, evaddr))
			!= ADDREG_OK) {
                    /* failed, return it */
		    return(return_val);
		}
                /* success, return ok */
		return(ADDREG_OK);
	    }
	}
        /* svaddr_old <= svaddr < evaddr_old */
	if (evaddr <= evaddr_old) {
	    /* be encircled by the original region, no need a region */
	    /* return ok */
	    return(MG_ADDREG_OK);
	} else {
	    /* overlay from right, truncate left portion */
	    svaddr = evaddr_old;
	    continue;
	}
    }
    /* the region has been adjusted */

    /*
     * 2. check free regions
     */
    if (mgIndexRegion + 1 > mgNregion) {
	/* no enough free regions, return error */
	return(ADDREG_FULL);
    }

    /*
     * 3. allocate free pages
     */
    /* calculate number of pages */
    npage = mgNbyte2Npage((u_int)evaddr - (u_int)svaddr);

    if ((pgp = (mgpage_t *)mgAllocate(npage, sizeof(mgpage_t)))
	    == NULL) {
        /* allocate failed, return error */
        return(ADDREG_NOSPACE);
    }
    /* success */

    /*
     * 4. allocate the region
     */
    mgIndexRegion ++;

    /*
     * 5. initialize the region
     */
    pregion = &mgRegions[mgIndexRegion - 1];
    pregion->gr_svaddr = svaddr;
    pregion->gr_evaddr = evaddr;
    pregion->gr_pages = pgp;
    pregion->gr_npage = npage;

    /*
     * 6. initialize pages
     */
    for (ix = 0; ix < npage; ix ++, pgp ++, svaddr += MG_SZPAGE) {
        /* initialize the page */
	/* the protection will be got when the first invariant is added
	mgGetInitProt(mgAddr2Pageno(mgVaddr2Paddr(svaddr)), &pgp->gp_prot);
	*/
	pgp->gp_invariants = NULL;
    }

    /*
     * 7. return ok
     */
    return(ADDREG_OK);
}

