
/*
 *  source - $Id: memguard_private.c,v 1.3 1997/06/10 18:49:27 qian Exp $
 *
 *  private functions for "memguard.c",
 *          a library for memory-based guard tool
 */

#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/swap.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/pgtable.h>
#include <asm/string.h>

#include <asm/memguard_machdep.h>
#include <memguard/macros.h>
#include <memguard/types.h>

#include "memguard_private.h"

#include <asm/memguard.h>
#include <memguard/internal.h>
#include <memguard/memguard_internal.h>


/*********************************************************************
variables (public)
*********************************************************************/


/*********************************************************************
variables (private)
*********************************************************************/


/*********************************************************************
function prototypes (private)
*********************************************************************/


/*********************************************************************
functions (public)
*********************************************************************/

/*
 * Function   : mg_enable
 * Description: enable the Tool's memguard page protection fault
 * Parameters : none
 * Returns    : none
 * Plan       : 1. search regions to change page protection to read-only
 * Note       : before this function, the tool should be disabled
 */
void
mg_enable()
{
    register mgregion_t *pregion;

    /*
     * 1. search regions to change page protection to read-only
     */
    for (pregion = &mgRegions[mgIndexRegion - 1];
	    pregion >= &mgRegions[0]; pregion --) {
        /* change protection of pages */
        register mgpage_t *pgp, *pgp_tail;
        caddr_t vaddr;

	for (pgp_tail = (pgp = pregion->gr_pages) + pregion->gr_npage,
		vaddr = pregion->gr_svaddr;
		pgp < pgp_tail; pgp ++, vaddr += MG_SZPAGE) {
            /* check if there is invariant in this page */
	    if (pgp->gp_invariants != NULL) {
	        /* there is invariant, change page protection */
	        mgSetReadOnly(mgAddr2Pageno(mgVaddr2Paddr(vaddr)),
		    &pgp->gp_prot);
	    }
        }
    }
}

/*
 * Function   : mg_disable
 * Description: disable the Tool's memguard page protection fault
 * Parameters : none
 * Returns    : none
 * Plan       : 1. search regions to change back page protection
 * Note       : before this function, the tool should be enabled
 */
void
mg_disable()
{
    register mgregion_t *pregion;

    /*
     * 1. search regions to change back page protection
     */
    for (pregion = &mgRegions[mgIndexRegion - 1];
	    pregion >= &mgRegions[0]; pregion --) {
        /* change protection of pages */
        register mgpage_t *pgp, *pgp_tail;
        caddr_t vaddr;

	for (pgp_tail = (pgp = pregion->gr_pages) + pregion->gr_npage,
		vaddr = pregion->gr_svaddr;
		pgp < pgp_tail; pgp ++, vaddr += MG_SZPAGE) {
            /* check if there is invariant in this page */
	    if (pgp->gp_invariants != NULL) {
	        /* there is invariant, change back page protection */
	        mgSetProt(mgAddr2Pageno(mgVaddr2Paddr(vaddr)),
		    pgp->gp_prot);
	    }
        }
    }
}

/*
 * Function   : mg_checkconflict
 * Description: check conflict
 * Parameters : pgp   - the start page pointer
 *              name  - invariant name
 *              vaddr - invariant virtual address
 *              size  - invariant size
 * Returns    : pointer to conflicting invariant (mginvariant_t *)
 *              == NULL	no conflict
 *              != NULL	conflict, pointer to conflicting invariant
 * Plan       : 1. check corresponding pages
 *              2. no conflict, return NULL
 * Note       : assume valid size
 */
mginvariant_t *
mg_checkconflict(mgpage_t *pgp, char *name, caddr_t vaddr, u_int size)
{
    register int offset;	/* offset in the page */
    register int length;	/* length in the page */
    register int size_left;	/* size left */
    mginvariant_t *pinvariant;
	/* pointer to the invariant control block */

    /*
     * 1. check corresponding pages
     */
    offset = mgAddr2Offset(vaddr);
    for (size_left = size; size_left > 0; size_left -= length) {
	int offset_end;	/* end offset */

        if ((length = MG_SZPAGE - offset) > size_left) {
	    length = size_left;
        }

        /* check page control block */
	offset_end = offset + length;
	for (pinvariant = pgp->gp_invariants; pinvariant != NULL;
		pinvariant = pinvariant->gi_next) {
            int offset_invariant;	/* offset of invariant */

	    /* check no conflict */
	    if (offset_end <= (offset_invariant = pinvariant->gi_offset)
	    || offset >= offset_invariant + pinvariant->gi_length) {
                /* no conflict, continue */
                continue;
            }
	    /* conflict */

            /* report error, return conflicting invariant */
	    mgReportConflict(pinvariant, name, vaddr, size);
	    return(pinvariant);
        }

	pgp ++;
	mgOffsetIncr(offset, length);
    }
    /* no conflict */

    /*
     * 2. no conflict, return NULL
     */
    return(NULL);
}

/*
 * Function   : mg_search
 * Description: search the invariant
 * Parameters : pgp   - the start page pointer
 *              vaddr - invariant virtual address
 *              size  - invariant size
 * Returns    : return flag (int)
 *              SEARCH_SUCCESS	found
 *              SEARCH_VADDR	error virtual address: non-exists
 *              SEARCH_SIZE	error size
 * Plan       : 1. search this page
 *              2. not found, report error, return error
 * Note       : assume valid size
 */
int
mg_search(mgpage_t *pgp, caddr_t vaddr, u_int size)
{
    register int offset;	/* offset in the page */
    mginvariant_t *pinvariant;
	/* pointer to the invariant control block */

    /*
     * 1. search this page
     */
    offset = mgAddr2Offset(vaddr);
    for (pinvariant = pgp->gp_invariants; pinvariant != NULL;
	    pinvariant = pinvariant->gi_next) {
	/* compare offset */
	if (offset != pinvariant->gi_offset) {
            /* offset not match, continue */
	    continue;
	}
	/* offset match, check invariant size */

	/* check if not the first part of an invariant */
	if (pinvariant->gi_size == 0) {
	    /* not the first part, break as non-exists */
	    break;
	}
	/* the first part of the invariant */

        /* check if care size and size do not match */
	if (size != 0 && size != pinvariant->gi_size) {
            /* care size and size do not match,
	       break -> report error, return error */
            mgReportSize(pinvariant, vaddr, size);
            return(SEARCH_SIZE);
        }
	/* do not care size or invariant size match */

	/* return ok */
	return(SEARCH_SUCCESS);
    }

    /*
     * 3. non-exists, report error, return error
     */
    mgReportNotfound(pinvariant, vaddr, size);
    return(SEARCH_VADDR);
}

/*
 * Function   : mg_add
 * Description: add the invariant
 * Parameters : pgp   - the start page pointer
 *              name  - invariant name
 *              vaddr - invariant virtual address
 *              size  - invariant size
 * Returns    : none
 * Plan       : 1. add to corresponding pages
 * Note       : assume no conflict
 *              assume enough free invariant control blocks
 *              assume valid size
 */
void
mg_add(mgpage_t *pgp, char *name, caddr_t vaddr, u_int size)
{
    register int offset;	/* offset in the page */
    register int length;	/* length in the page */
    register int size_left;	/* size left */
    mginvariant_t *pinvariant;
	/* pointer to the invariant control block */

    int bfirstloop = 1;		/* boolean for first loop */

    /*
     * 1. add to corresponding pages
     */
    offset = mgAddr2Offset(vaddr);
    for (size_left = size; size_left > 0; size_left -= length) {
        if ((length = MG_SZPAGE - offset) > size_left) {
	    length = size_left;
        }

        /* allocate a free invariant control block */
        mgInvariantsFree = (pinvariant = mgInvariantsFree)->gi_next;

        /* initialize it */
	strncpy(pinvariant->gi_name, name, MG_SZNAME);
        pinvariant->gi_offset = offset;
	pinvariant->gi_length = length;
	if (bfirstloop) {
	    /* the first part of invariant */
	    bfirstloop = 0;
	    pinvariant->gi_size = size;
        } else {
	    /* not the first part of invariant */
	    pinvariant->gi_size = 0;
	}

        /* add it, may change page protection */
	if ((pinvariant->gi_next = pgp->gp_invariants) == NULL) {
	    /* may change page protection */
	    if (mgEnabled) {
#ifdef MG_DEBUG
		printk("mg_add: mgSetReadOnly(pageno=%x, &%x)\n",
			(int)mgAddr2Pageno(mgVaddr2Paddr(vaddr+size-size_left)),
			(int)(pgp->gp_prot));
#endif
	        mgSetReadOnly(mgAddr2Pageno(
		    mgVaddr2Paddr(vaddr + size - size_left)),
		    &pgp->gp_prot);
            }
	}
	pgp->gp_invariants = pinvariant;

	pgp ++;
	mgOffsetIncr(offset, length);
    }
    /* loop finish */
}

/*
 * Function   : mg_delete
 * Description: delete the invariant
 * Parameters : pgp   - the start page pointer
 *              vaddr - invariant virtual address
 * Returns    : return flag (int)
 *              DELETE_SUCCESS	found
 *              DELETE_VADDR	error virtual address: non-exists
 *              DELETE_SIZE	error size
 * Plan       : 1. get the first invariant
 *              2. delete from corresponding pages
 *              3. return ok
 * Note       : none
 */
int
mg_delete(mgpage_t *pgp, caddr_t vaddr)
{
    register int offset;	/* offset in the page */
    register int length;	/* length in the page */
    register int size_left;	/* size left */
    u_int        size;		/* size of invariant */
				/* NOTE: ignore the warning message from gcc */
    mginvariant_t *pinvariant;
	/* pointer to the invariant control block */

    /*
     * 1. get the first invariant
     */
    offset = mgAddr2Offset(vaddr);
    for (pinvariant = pgp->gp_invariants; pinvariant != NULL;
	    pinvariant = pinvariant->gi_next) {
        if (pinvariant->gi_offset == offset) {
	    if ((size = pinvariant->gi_size) == 0) {
                /* the middle of a quasi-invariant */
                return(DELETE_VADDR);
            }
	    break;
	}
    }
    /* check if not found */
    if (pinvariant == NULL) {
        return(DELETE_VADDR);
    }

    /*
     * 2. delete from corresponding pages
     */
    for (size_left = size; size_left > 0; size_left -= length) {
	mginvariant_t *pinvariant_prev;
	    /* pointer to the previous one */

        if ((length = MG_SZPAGE - offset) > size_left) {
	    length = size_left;
        }

        /* delete from page control block */
	for (pinvariant_prev = NULL, pinvariant = pgp->gp_invariants;
		pinvariant != NULL;
		pinvariant_prev = pinvariant,
		    pinvariant = pinvariant->gi_next) {
	    /* compare offset */
	    if (offset != pinvariant->gi_offset) {
                /* offsets mis-match */
                continue;
            }
            /* offsets match */

            /* delete it, may change back page protection */
            if (pinvariant_prev == NULL) {
		/* the head of the chain,
		   delete it, may change back page protection */
		if ((pgp->gp_invariants = pinvariant->gi_next) == NULL) {
		    /* may change back page protection */
		    if (mgEnabled) {
	                mgSetProt(mgAddr2Pageno(
			    mgVaddr2Paddr(vaddr + size - size_left)),
		            pgp->gp_prot);
                    }
                }
	    } else {
		/* not the head of the chain */
		pinvariant_prev->gi_next = pinvariant->gi_next;
	    }

	    /* free it */
	    pinvariant->gi_next = mgInvariantsFree;
            mgInvariantsFree = pinvariant;

	    /* break, go to next page */
	    break;
        }
	/* finish this page */

	pgp ++;
	mgOffsetIncr(offset, length);
    }
    /* loop finish */

    /*
     * 3. return ok
     */
    return(DELETE_SUCCESS);
}

/*
 * Function   : mg_sga
 * Description: set/get/assert the invariant
 * Parameters : flag    - set/get/assert
 *              pgp     - the start page pointer
 *              vaddr   - invariant virtual address
 *              size    - invariant size
 *              content - invariant content
 *                  set in / get out / assert in
 * Returns    : return flag (int)
 *              SGA_OK		success
 *              SGA_ERROR	error
 * Plan       : 1. process corresponding pages
 *              2. return ok
 * Note       : assume exists
 *              assume valid size
 */
int
mg_sga(int flag, mgpage_t *pgp,
       caddr_t vaddr, u_int size, char *content)
{
    register int offset;	/* offset in the page */
    register int length;	/* length in the page */
    register int size_left;	/* size left */
    mginvariant_t *pinvariant;
	/* pointer to the invariant control block */

    static char TmpPage[MG_SZPAGE];

    /*
     * 1. process corresponding pages
     */
    offset = mgAddr2Offset(vaddr);
    for (size_left = size; size_left > 0; size_left -= length) {
        if ((length = MG_SZPAGE - offset) > size_left) {
	    length = size_left;
        }

        /* process page control block */
	for (pinvariant = pgp->gp_invariants; pinvariant != NULL;
		pinvariant = pinvariant->gi_next) {
	    /* compare offset */
	    if (offset != pinvariant->gi_offset) {
                /* offset not match, continue */
	        continue;
	    }
	    /* offset match */
	    
	    /* do something */
            switch (flag) {
	    case SGA_SET :	/* set */
		/* write to physical page */
		if (mgWrite(content, mgAddr2Pageno(
			mgVaddr2Paddr(vaddr + size - size_left)),
			offset, length) != length) {
                    /* write error, return error */
		    return(SGA_ERROR);
                }
		break;

	    case SGA_GET :	/* get */
		/* read from physical page */
		if (mgRead(content, mgAddr2Pageno(
			mgVaddr2Paddr(vaddr + size - size_left)),
			offset, length) != length) {
                    /* read error, return error */
		    return(SGA_ERROR);
                }
		break;

	    case SGA_ASSERT :	/* assert */
		/* read from physical page to TmpPage */
		if (mgRead(TmpPage, mgAddr2Pageno(
			mgVaddr2Paddr(vaddr + size - size_left)),
			offset, length) != length) {
                    /* read error, return error */
		    return(SGA_ERROR);
                }

		/* compare TmpPage and content */
                /* for HPUX: define memcmp to  bcmp */
                if (memcmp(TmpPage, content, length) != 0) {
		    /* not match, return error */
		    return(SGA_ERROR);
		}

		/* match */
		break;
	    }
	    /* finish this page */

	    /* break, go to next page */
	    break;
        }

        content += length;
	pgp ++;
	mgOffsetIncr(offset, length);
    }
    /* loop finish */

    /*
     * 3. return ok
     */
    return(SGA_OK);
}


/*********************************************************************
functions (private)
*********************************************************************/

